;;; init.el --- bandali's emacs configuration -*- lexical-binding: t -*-

;; Copyright (C) 2018-2020  Amin Bandali <bandali@gnu.org>

;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; GNU Emacs configuration of Amin Bandali, computer scientist,
;; Free Software activist, and GNU maintainer & webmaster.  Packages
;; are installed through using Borg for a fully reproducible setup.

;; Over the years, I've taken inspiration from configurations of many
;; great people.  Some that I can remember off the top of my head are:
;;
;; - https://github.com/dieggsy/dotfiles
;; - https://github.com/dakra/dmacs
;; - http://pages.sachachua.com/.emacs.d/Sacha.html
;; - https://github.com/dakrone/eos
;; - http://doc.rix.si/cce/cce.html
;; - https://github.com/jwiegley/dot-emacs
;; - https://github.com/wasamasa/dotemacs
;; - https://github.com/hlissner/doom-emacs

;;; Code:

;;; Emacs initialization

(defvar b/before-user-init-time (current-time)
  "Value of `current-time' when Emacs begins loading `user-init-file'.")
(defvar b/emacs-initialized nil
  "Whether Emacs has been initialized.")
(defvar b/exwm-p (string= (system-name) "chaman")
  "Whether or not we will be using `exwm'.")

(when (not (bound-and-true-p b/emacs-initialized))
  (message "Loading Emacs...done (%.3fs)"
           (float-time (time-subtract b/before-user-init-time
                                      before-init-time))))

;; temporarily increase `gc-cons-threshhold' and `gc-cons-percentage'
;; during startup to reduce garbage collection frequency.  clearing
;; `file-name-handler-alist' seems to help reduce startup time too.
(defvar b/gc-cons-threshold gc-cons-threshold)
(defvar b/gc-cons-percentage gc-cons-percentage)
(defvar b/file-name-handler-alist file-name-handler-alist)
(setq gc-cons-threshold (* 30 1024 1024)  ; 30 MiB
      gc-cons-percentage 0.6
      file-name-handler-alist nil
      ;; sidesteps a bug when profiling with esup
      esup-child-profile-require-level 0)

;; set them back to their defaults once we're done initializing
(defun b/post-init ()
  "My post-initialize function, run after loading `user-init-file'."
  (setq b/emacs-initialized     t
        gc-cons-threshold       b/gc-cons-threshold
        gc-cons-percentage      b/gc-cons-percentage
        file-name-handler-alist b/file-name-handler-alist)
  (when b/exwm-p
    (with-eval-after-load 'exwm-workspace
      (setq-default
       mode-line-format
       (append
        mode-line-format
        '((:eval
           (format
            "[%s]" (number-to-string
                    exwm-workspace-current-index)))))))))
(add-hook 'after-init-hook #'b/post-init)

;; increase number of lines kept in *Messages* log
(setq message-log-max 20000)

;; optionally, uncomment to supress some byte-compiler warnings
;;   (see C-h v byte-compile-warnings RET for more info)
;; (setq byte-compile-warnings
;;       '(not free-vars unresolved noruntime lexical make-local))


;;; whoami

(setq user-full-name "Amin Bandali"
      user-mail-address "bandali@gnu.org")


;;; comment macro

;; useful for commenting out multiple sexps at a time
(defmacro comment (&rest _)
  "Comment out one or more s-expressions."
  (declare (indent defun))
  nil)


;;; Package management

(progn ;   `borg'
  (add-to-list 'load-path
               (expand-file-name "lib/borg" user-emacs-directory))
  (require 'borg)
  (borg-initialize)
  (setq borg-rewrite-urls-alist
        '(("git@github.com:" . "https://github.com/")
          ("git@gitlab.com:" . "https://gitlab.com/"))))

;; use-package
(if nil                             ; set to t when need to debug init
    (progn
      (setq use-package-verbose t
            use-package-expand-minimally nil
            use-package-compute-statistics t
            debug-on-error t)
      (require 'use-package))
  (setq use-package-verbose nil
        use-package-expand-minimally t))

(setq use-package-always-defer t)
(require 'bind-key)


;;; Initial setup

;; keep ~/.emacs.d clean
(use-package no-littering
  :demand
  :config
  (defalias 'b/etc 'no-littering-expand-etc-file-name)
  (defalias 'b/var 'no-littering-expand-var-file-name))

(use-package auto-compile
  :demand
  :config
  (auto-compile-on-load-mode)
  (auto-compile-on-save-mode)
  (setq auto-compile-display-buffer               nil)
  (setq auto-compile-mode-line-counter            t)
  (setq auto-compile-source-recreate-deletes-dest t)
  (setq auto-compile-toggle-deletes-nonlib-dest   t)
  (setq auto-compile-update-autoloads             t))

;; separate custom file (don't want it mixing with init.el)
(use-package custom
  :no-require
  :config
  (setq custom-file (b/etc "custom.el"))
  (when (file-exists-p custom-file)
    (load custom-file))
  ;; while at it, treat themes as safe
  (setf custom-safe-themes t)
  ;; only one custom theme at a time
  (comment
    (defadvice load-theme (before clear-previous-themes activate)
      "Clear existing theme settings instead of layering them"
      (mapc #'disable-theme custom-enabled-themes))))

;; load the secrets file if it exists, otherwise show a warning
(comment
  (with-demoted-errors
      (load (b/etc "secrets"))))

;; start up emacs server.  see
;; https://www.gnu.org/software/emacs/manual/html_node/emacs/Emacs-Server.html#Emacs-Server
(use-package server
  :defer 0.5
  :config
  (declare-function server-edit "server")
  (bind-key "C-c F D" 'server-edit)
  (declare-function server-running-p "server")
  (or (server-running-p) (server-mode)))


;;; Useful utilities

(defmacro b/setq-every (value &rest vars)
  "Set all the variables from VARS to value VALUE."
  (declare (indent defun) (debug t))
  `(progn ,@(mapcar (lambda (x) (list 'setq x value)) vars)))

(defun b/start-process (program &rest args)
  "Same as `start-process', but doesn't bother about name and buffer."
  (let ((process-name (concat program "_process"))
        (buffer-name  (generate-new-buffer-name
                       (concat program "_output"))))
    (apply #'start-process
           process-name buffer-name program args)))

(defun b/dired-start-process (program &optional args)
  "Open current file with a PROGRAM."
  ;; Shell command looks like this: "program [ARGS]... FILE" (ARGS can
  ;; be nil, so remove it).
  (declare-function dired-get-file-for-visit "dired")
  (apply #'b/start-process
         program
         (remove nil (list args (dired-get-file-for-visit)))))

(defun b/add-elisp-section ()
  (interactive)
  (insert "\n")
  (forward-line -1)
  (insert "\n\n;;; "))

;; (defvar b/fill-column 47
;;   "My custom `fill-column'.")

(defconst b/asterism "* * *")

(defun b/insert-asterism ()
  "Insert a centred asterism."
  (interactive)
  (insert
   (concat
    "\n\n"
    (make-string (floor (/ (- fill-column (length b/asterism)) 2))
                 ?\s)
    b/asterism
    "\n\n")))

(defun b/no-mouse-autoselect-window ()
  "Conveniently disable `focus-follows-mouse'.
For disabling the behaviour for certain buffers and/or modes."
  (make-local-variable 'mouse-autoselect-window)
  (setq mouse-autoselect-window nil))

(defun b/kill-current-buffer ()
  "Kill the current buffer."
  ;; also see https://redd.it/64xb3q
  (interactive)
  (kill-buffer (current-buffer)))

(defun b/move-indentation-or-beginning-of-line (arg)
  "Move to the indentation or to the beginning of line."
  (interactive "^p")
  ;; (if (bolp)
  ;;     (back-to-indentation)
  ;;   (move-beginning-of-line arg))
  (if (= (point)
         (progn (back-to-indentation)
                (point)))
      (move-beginning-of-line arg)))

(defun b/join-line-top ()
  "Like `join-line', but join next line to the current line."
  (interactive)
  (join-line 1))

(defun b/duplicate-line-or-region (&optional n)
  "Duplicate the current line, or region (if active).
Make N (default: 1) copies of the current line or region."
  (interactive "*p")
  (let ((u-r-p (use-region-p))          ; if region is active
        (n1 (or n 1)))
    (save-excursion
      (let ((text
             (if u-r-p
                 (buffer-substring (region-beginning) (region-end))
               (prog1 (thing-at-point 'line)
                 (end-of-line)
                 (if (eobp)
                     (newline)
                   (forward-line 1))))))
        (dotimes (_ (abs n1))
          (insert text))))
    (let* ((beg (if u-r-p
                    (region-beginning)
                  (line-beginning-position)))
           (end (if u-r-p
                    (region-end)
                  (line-end-position)))))))


;;; Defaults

;;;; C-level customizations

(setq
 ;; minibuffer
 enable-recursive-minibuffers t
 resize-mini-windows t
 ;; more useful frame titles
 frame-title-format '("" invocation-name " - "
                      (:eval
                       (if (buffer-file-name)
                           (abbreviate-file-name (buffer-file-name))
                         "%b")))
 ;; i don't feel like jumping out of my chair every now and again; so
 ;; don't BEEP! at me, emacs
 ring-bell-function 'ignore
 ;; better scrolling
 ;; scroll-margin 1
 ;; scroll-conservatively 10000
 scroll-step 1
 scroll-conservatively 101
 scroll-preserve-screen-position 1
 ;; focus follows mouse
 mouse-autoselect-window t)

(setq-default
 ;; always use space for indentation
 indent-tabs-mode nil
 tab-width 4
 ;; cursor shape
 cursor-type t)

;; unicode support
(comment
  (dolist (ft (fontset-list))
    (set-fontset-font
     ft
     'unicode
     (font-spec :name "Source Code Pro" :size 14))
    (set-fontset-font
     ft
     'unicode
     (font-spec :name "DejaVu Sans Mono")
     nil
     'append)
    ;; (set-fontset-font
    ;;  ft
    ;;  'unicode
    ;;  (font-spec
    ;;   :name "Symbola monospacified for DejaVu Sans Mono")
    ;;  nil
    ;;  'append)
    ;; (set-fontset-font
    ;;  ft
    ;;  #x2115  ; ℕ
    ;;  (font-spec :name "DejaVu Sans Mono")
    ;;  nil
    ;;  'append)
    (set-fontset-font
     ft
     (cons ?Α ?ω)
     (font-spec :name "DejaVu Sans Mono" :size 14)
     nil
     'prepend)))

;;;; Elisp-level customizations

(use-package startup
  :no-require
  :demand
  :config
  ;; don't need to see the startup echo area message
  (advice-add #'display-startup-echo-area-message :override #'ignore)
  :custom
  ;; i want *scratch* as my startup buffer
  (initial-buffer-choice t)
  ;; i don't need the default hint
  (initial-scratch-message nil)
  ;; use customizable text-mode as major mode for *scratch*
  ;; (initial-major-mode 'text-mode)
  ;; inhibit buffer list when more than 2 files are loaded
  (inhibit-startup-buffer-menu t)
  ;; don't need to see the startup screen or echo area message
  (inhibit-startup-screen t)
  (inhibit-startup-echo-area-message user-login-name))

(use-package files
  :no-require
  :demand
  :custom
  ;; backups (C-h v make-backup-files RET)
  (backup-by-copying t)
  (version-control t)
  (delete-old-versions t)

  ;; auto-save
  (auto-save-file-name-transforms
   `((".*" ,(b/var "auto-save/") t)))

  ;; insert newline at the end of files
  (require-final-newline t)

  ;; open read-only file buffers in view-mode
  ;; (enables niceties like `q' for quit)
  (view-read-only t))

;; disable disabled commands
(setq disabled-command-function nil)

;; lazy-person-friendly yes/no prompts
(defalias 'yes-or-no-p #'y-or-n-p)

;; enable automatic reloading of changed buffers and files
(use-package autorevert
  :demand
  :config
  (global-auto-revert-mode 1)
  :custom
  (auto-revert-verbose nil)
  (global-auto-revert-non-file-buffers nil))

;; time and battery in mode-line
(use-package time
  :demand
  :config
  (display-time-mode)
  :custom
  (display-time-default-load-average nil)
  (display-time-format " %a %b %-e %-l:%M%P")
  (display-time-mail-icon '(image :type xpm :file "gnus/gnus-pointer.xpm" :ascent center))
  (display-time-use-mail-icon t))

(use-package battery
  :demand
  :config
  (display-battery-mode)
  :custom
  (battery-mode-line-format "%p%% %t"))

(use-package fringe
  :demand
  :config
  ;; smaller fringe
  ;; (fringe-mode '(3 . 1))
  (fringe-mode nil))

(use-package winner
  :demand
  :config
  ;; enable winner-mode (C-h f winner-mode RET)
  (winner-mode 1))

(use-package compile
  :config
  ;; don't display *compilation* buffer on success.  based on
  ;; https://stackoverflow.com/a/17788551, with changes to use `cl-letf'
  ;; instead of the now obsolete `flet'.
  (defun b/compilation-finish-function (buffer outstr)
    (unless (string-match "finished" outstr)
      (switch-to-buffer-other-window buffer))
    t)

  (setq compilation-finish-functions #'b/compilation-finish-function)

  (require 'cl-macs)

  (defadvice compilation-start
      (around inhibit-display
              (command &optional mode name-function highlight-regexp))
    (if (not (string-match "^\\(find\\|grep\\)" command))
        (cl-letf (((symbol-function 'display-buffer) #'ignore))
          (save-window-excursion ad-do-it))
      ad-do-it))
  (ad-activate 'compilation-start))

(use-package isearch
  :custom
  ;; allow scrolling in Isearch
  (isearch-allow-scroll t)
  ;; search for non-ASCII characters: i’d like non-ASCII characters such
  ;; as ‘’“”«»‹›áⓐ𝒶 to be selected when i search for their ASCII
  ;; counterpart.  shoutout to
  ;; http://endlessparentheses.com/new-in-emacs-25-1-easily-search-non-ascii-characters.html
  (search-default-mode #'char-fold-to-regexp))

;; uncomment to extend the above behaviour to query-replace
(comment
  (use-package replace
    :custom
    (replace-char-fold t)))

(use-package vc
  :bind ("C-x v C-=" . vc-ediff))

(use-package vc-git
  :after vc
  :custom
  (vc-git-print-log-follow t))

(use-package ediff
  :config (add-hook 'ediff-after-quit-hook-internal 'winner-undo)
  :custom ((ediff-window-setup-function 'ediff-setup-windows-plain)
           (ediff-split-window-function 'split-window-horizontally)))

(use-package face-remap
  :custom
  ;; gentler font resizing
  (text-scale-mode-step 1.05))

(use-package mwheel
  :defer 0.4
  :config
  (setq mouse-wheel-scroll-amount '(1 ((shift) . 1)) ; one line at a time
        mouse-wheel-progressive-speed nil            ; don't accelerate scrolling
        mouse-wheel-follow-mouse t))                 ; scroll window under mouse

(use-package pixel-scroll
  :defer 0.4
  :config (pixel-scroll-mode 1))

(use-package epg-config
  :config
  ;; ask for GPG passphrase in minibuffer
  ;; this will fail if gpg>=2.1 is not available
  (setq epg-pinentry-mode 'loopback)
  :custom
  (epg-gpg-program (executable-find "gpg")))

(use-package epg
  :after epg-config)

(use-package pinentry
  :disabled
  :demand
  :after (epa epg server)
  :config
  ;; workaround for systemd-based distros:
  ;; (setq pinentry--socket-dir server-socket-dir)
  (pinentry-start))

(use-package auth-source
  :custom
  (auth-sources '("~/.authinfo.gpg"))
  (authinfo-hidden (regexp-opt '("password" "client-secret" "token"))))


;;; General bindings

(bind-keys
 ("C-a"     . b/move-indentation-or-beginning-of-line)
 ("C-c a i" . ielm)
 ("C-c d"   . b/duplicate-line-or-region)

 ("C-c e b" . eval-buffer)
 ("C-c e e" . eval-last-sexp)
 ("C-c e p" . pp-macroexpand-last-sexp)
 ("C-c e r" . eval-region)

 ("C-c e i" . emacs-init-time)
 ("C-c e u" . emacs-uptime)
 ("C-c e v" . emacs-version)

 ("C-c f ." . find-file)
 ("C-c f d" . find-name-dired)
 ("C-c f l" . find-library)

 ("C-c F m" . make-frame-command)
 ("C-c F d" . delete-frame)

 ("C-S-h C" . describe-char)
 ("C-S-h F" . describe-face)

 ("C-S-j"   . b/join-line-top)

 ("C-c x"   . execute-extended-command)

 ("C-x k"   . b/kill-current-buffer)
 ("C-x K"   . kill-buffer)
 ("C-x s"   . save-buffer)
 ("C-x S"   . save-some-buffers)

 :map emacs-lisp-mode-map
 ("<C-return>" . b/add-elisp-section))

(when (display-graphic-p)
  (unbind-key "C-z" global-map))

(bind-keys
 ;; for back and forward mouse keys
 ("<XF86Back>"     . previous-buffer)
 ("<mouse-8>"      . previous-buffer)
 ;; ("<drag-mouse-8>" . previous-buffer)
 ("<XF86Forward>"  . next-buffer)
 ("<mouse-9>"      . next-buffer)
 ;; ("<drag-mouse-9>" . next-buffer)
 ;; ("<drag-mouse-2>" . kill-this-buffer)
 ;; ("<drag-mouse-3>" . switch-to-buffer)
 )


;;; Essential packages

(add-to-list
 'load-path
 (expand-file-name
  (convert-standard-filename "lisp") user-emacs-directory))

(when b/exwm-p
  (require 'bandali-exwm))

(require 'bandali-org)

;; *the* right way to do git
(use-package magit
  :bind (("C-x g"   . magit-status)
         ("C-c g g" . magit-status)
         ("C-c g b" . magit-blame-addition)
         ("C-c g l" . magit-log-buffer-file))
  :config
  (declare-function magit-add-section-hook "magit-section"
                    (hook function &optional at append local))
  (magit-add-section-hook 'magit-status-sections-hook
                          'magit-insert-modules
                          'magit-insert-stashes
                          'append)
  ;; (magit-add-section-hook 'magit-status-sections-hook
  ;;                         'magit-insert-ignored-files
  ;;                         'magit-insert-untracked-files
  ;;                         'append)
  (setq magit-repository-directories '(("~/.emacs.d/" . 0)
                                       ("~/src/git/" . 2)))
  (nconc magit-section-initial-visibility-alist
         '(([unpulled status] . show)
           ([unpushed status] . show)))
  (declare-function magit-display-buffer-fullframe-status-v1 "magit-mode" (buffer))
  :custom
  (magit-diff-refine-hunk t)
  (magit-display-buffer-function #'magit-display-buffer-fullframe-status-v1)
  ;; (magit-completing-read-function 'magit-ido-completing-read)
  :custom-face (magit-diff-file-heading ((t (:weight normal)))))

;; recently opened files
(use-package recentf
  :defer 0.2
  ;; :config
  ;; (add-to-list 'recentf-exclude "^/\\(?:ssh\\|su\\|sudo\\)?:")
  :config
  (recentf-mode)
  :custom
  (recentf-max-saved-items 2000))

;; needed for history for counsel
(use-package amx
  :defer 0.3
  :config
  (amx-mode))

;; (require 'bandali-ido)
(require 'bandali-ivy)

(require 'bandali-eshell)

(require 'bandali-ibuffer)

(use-package outline
  :disabled
  :hook (prog-mode . outline-minor-mode)
  :bind
  (:map
   outline-minor-mode-map
   ("<s-tab>"  . outline-toggle-children)
   ("M-p"      . outline-previous-visible-heading)
   ("M-n"      . outline-next-visible-heading)
   :prefix-map b/outline-prefix-map
   :prefix "s-O"
   ("TAB" . outline-toggle-children)
   ("a"   . outline-hide-body)
   ("H"   . outline-hide-body)
   ("S"   . outline-show-all)
   ("h"   . outline-hide-subtree)
   ("s"   . outline-show-subtree)))

(use-package ls-lisp
  :custom (ls-lisp-dirs-first t))

(require 'bandali-dired)

(use-package help
  :config
  (temp-buffer-resize-mode)
  (setq help-window-select t))

(use-package tramp
  :config
  (add-to-list 'tramp-default-proxies-alist '(nil "\\`root\\'" "/ssh:%h:"))
  (add-to-list 'tramp-default-proxies-alist '("localhost" nil nil))
  (add-to-list 'tramp-default-proxies-alist
               (list (regexp-quote (system-name)) nil nil)))

(use-package doc-view
  :bind (:map doc-view-mode-map
              ("M-RET" . image-previous-line)))

;; Email (with Gnus, message, and EBDB)
(require 'bandali-gnus)
(use-package sendmail
  :config
  (setq sendmail-program (executable-find "msmtp")
        ;; message-sendmail-extra-arguments '("-v" "-d")
        mail-specify-envelope-from t
        mail-envelope-from 'header))
(require 'bandali-message)
(require 'bandali-ebdb)

;; IRC (with ERC and ZNC)
(require 'bandali-erc)

(use-package scpaste
  :config
  (setq scpaste-http-destination "https://p.bndl.org"
        scpaste-scp-destination "p:~"))


;;; Editing

;; highlight uncommitted changes in the left fringe
(use-package diff-hl
  :defer 0.6
  :config
  (setq diff-hl-draw-borders nil)
  (global-diff-hl-mode)
  :hook
  ((magit-pre-refresh . diff-hl-magit-pre-refresh)
   (magit-post-refresh . diff-hl-magit-post-refresh)))

;; display Lisp objects at point in the echo area
(use-package eldoc
  :when (version< "25" emacs-version)
  :config (global-eldoc-mode))

;; highlight matching parens
(use-package paren
  :demand
  :config (show-paren-mode))

(use-package elec-pair
  :demand
  :config (electric-pair-mode))

(use-package simple
  :config (column-number-mode)
  :custom
  ;; Save what I copy into clipboard from other applications into Emacs'
  ;; kill-ring, which would allow me to still be able to easily access
  ;; it in case I kill (cut or copy) something else inside Emacs before
  ;; yanking (pasting) what I'd originally intended to.
  (save-interprogram-paste-before-kill t))

;; save minibuffer history
(use-package savehist
  :demand
  :config
  (savehist-mode)
  (add-to-list 'savehist-additional-variables 'kill-ring))

;; automatically save place in files
(use-package saveplace
  :when (version< "25" emacs-version)
  :config (save-place-mode))

(use-package prog-mode
  :config (global-prettify-symbols-mode)
  (defun indicate-buffer-boundaries-left ()
    (setq indicate-buffer-boundaries 'left))
  (add-hook 'prog-mode-hook #'indicate-buffer-boundaries-left))

(use-package text-mode
  :bind (:map text-mode-map ("C-*" . b/insert-asterism))
  :hook ((text-mode . indicate-buffer-boundaries-left)
         (text-mode . flyspell-mode)))

(use-package conf-mode
  :mode "\\.*rc$")

(use-package sh-script
  :mode ("\\.bashrc$" . sh-mode))

(use-package company
  :disabled
  :bind
  (:map company-active-map
        ([tab]    . company-complete-common-or-cycle)
        ([escape] . company-abort)
        ("C-p"    . company-select-previous-or-abort)
        ("C-n"    . company-select-next-or-abort))
  :custom
  (company-minimum-prefix-length 1)
  (company-selection-wrap-around t)
  (company-dabbrev-char-regexp "\\sw\\|\\s_\\|[-_]")
  (company-dabbrev-downcase nil)
  (company-dabbrev-ignore-case nil)
  ;; :config
  ;; (global-company-mode t)
  )

(use-package flycheck
  :disabled
  :defer 0.6
  :hook (prog-mode . flycheck-mode)
  :bind
  (:map flycheck-mode-map
        ("M-P" . flycheck-previous-error)
        ("M-N" . flycheck-next-error))
  :config
  ;; Use the load-path from running Emacs when checking elisp files
  (setq flycheck-emacs-lisp-load-path 'inherit)

  ;; Only flycheck when I actually save the buffer
  (setq flycheck-check-syntax-automatically '(mode-enabled save))
  :custom (flycheck-mode-line-prefix "flyc"))

;; (use-package flyspell)

;; http://endlessparentheses.com/ispell-and-apostrophes.html
(use-package ispell
  :disabled
  :defer 0.6
  :config
  ;; ’ can be part of a word
  (setq ispell-local-dictionary-alist
        `((nil "[[:alpha:]]" "[^[:alpha:]]"
               "['\x2019]" nil ("-B") nil utf-8))
        ispell-program-name (executable-find "hunspell"))
  ;; don't send ’ to the subprocess
  (defun endless/replace-apostrophe (args)
    (cons (replace-regexp-in-string
           "’" "'" (car args))
          (cdr args)))
  (advice-add #'ispell-send-string :filter-args
              #'endless/replace-apostrophe)

  ;; convert ' back to ’ from the subprocess
  (defun endless/replace-quote (args)
    (if (not (derived-mode-p 'org-mode))
        args
      (cons (replace-regexp-in-string
             "'" "’" (car args))
            (cdr args))))
  (advice-add #'ispell-parse-output :filter-args
              #'endless/replace-quote))

(use-package abbrev
  :hook (text-mode . abbrev-mode))


;;; Programming modes

(use-package lisp-mode
  :config
  (defun indent-spaces-mode ()
    (setq indent-tabs-mode nil))
  (add-hook 'lisp-interaction-mode-hook #'indent-spaces-mode))

(use-package alloy-mode
  :mode "\\.\\(als\\|dsh\\)\\'"
  :config
  (setq alloy-basic-offset 2)
  ;; (defun b/alloy-simple-indent (start end)
  ;;   (interactive "r")
  ;;   ;; (if (region-active-p)
  ;;   ;;     (indent-rigidly start end alloy-basic-offset)
  ;;   ;;   (if (bolp)
  ;;   ;;       (indent-rigidly (line-beginning-position)
  ;;   ;;                       (line-end-position)
  ;;   ;;                       alloy-basic-offset)))
  ;;   (indent-to (+ (current-column) alloy-basic-offset)))
  :bind (:map alloy-mode-map
              ("RET" . electric-newline-and-maybe-indent)
              ;; ("TAB" . b/alloy-simple-indent)
              ("TAB" . indent-for-tab-command))
  :hook (alloy-mode . (lambda () (setq-local indent-tabs-mode nil))))

(use-package lean-mode
  :disabled
  :defer 0.4
  :init (eval-when-compile (defvar lean-mode-map))
  :bind (:map lean-mode-map
              ("S-SPC" . company-complete))
  :config
  (require 'lean-input)
  (setq default-input-method "Lean"
        lean-input-tweak-all '(lean-input-compose
                               (lean-input-prepend "/")
                               (lean-input-nonempty))
        lean-input-user-translations '(("/" "/")))
  (lean-input-setup))

(use-package sgml-mode
  :config
  (setq sgml-basic-offset 0))

(use-package css-mode
  :config
  (setq css-indent-offset 2))

(use-package emmet-mode
  :after (:any mhtml-mode css-mode sgml-mode)
  :bind* (("C-)" . emmet-next-edit-point)
          ("C-(" . emmet-prev-edit-point))
  :config
  (unbind-key "C-j" emmet-mode-keymap)
  (setq emmet-move-cursor-between-quotes t)
  :hook (css-mode html-mode sgml-mode))

(use-package geiser
  :disabled)

(use-package geiser-guile
  :disabled
  :config
  (setq geiser-guile-load-path "~/src/git/guix"))

(use-package guix
  :disabled)

(use-package go-mode
  :disabled)

(use-package po-mode
  :disabled
  :hook
  (po-mode . (lambda () (run-with-timer 0.1 nil 'View-exit))))

(use-package auctex
  :disabled
  :custom
  (font-latex-fontify-sectioning 'color))

(use-package tex-mode
  :config
  (cl-delete-if
   (lambda (p) (string-match "^---?" (car p)))
   tex--prettify-symbols-alist)
  :hook ((tex-mode . auto-fill-mode)
         (tex-mode . flyspell-mode)))

;; (use-package george-mode
;;   :straight (:host nil :repo "https://git.shemshak.org/amin/george-mode")
;;   :mode "\\.grg\\'")


;;; Theme

(comment
(add-to-list 'custom-theme-load-path
             (expand-file-name
              (convert-standard-filename "lisp") user-emacs-directory))
(load-theme 'tangomod t)

(use-package smart-mode-line
  :commands (sml/apply-theme)
  :demand
  :config
  ;; thanks, but no thnaks; don't make fixed-width fills.
  (defun sml/fill-for-buffer-identification () "")
  (setq sml/theme 'tangomod)
  (sml/setup)
  (smart-mode-line-enable))

(use-package doom-modeline
  :disabled
  :demand
  :hook (after-init . doom-modeline-init)
  :custom
  (doom-modeline-buffer-file-name-style 'relative-to-project))

(use-package doom-themes)

(use-package moody
  :disabled
  :demand
  :config
  (setq x-underline-at-descent-line t)
  (let ((line (face-attribute 'mode-line :underline)))
    (set-face-attribute 'mode-line          nil :overline   line)
    (set-face-attribute 'mode-line-inactive nil :overline   line)
    (set-face-attribute 'mode-line-inactive nil :underline  line)
    (set-face-attribute 'mode-line          nil :box        nil)
    (set-face-attribute 'mode-line-inactive nil :box        nil)
    (set-face-attribute 'mode-line-inactive nil :background "#e1e1e1")) ; d3d7cf
  (moody-replace-mode-line-buffer-identification)
  (moody-replace-vc-mode))

(use-package mini-modeline
  :disabled
  :demand
  :config (mini-modeline-mode))

(defvar b/org-mode-font-lock-keywords
  '(("[ \t]*\\(#\\+\\(BEGIN\\|END\\|begin\\|end\\)_\\(\\S-+\\)\\)[ \t]*\\([^\n:]*\\)"
      (1 '(:foreground "#5a5b5a" :background "#292b2b") t) ; directive
      (3 '(:foreground "#81a2be" :background "#292b2b") t) ; kind
      (4 '(:foreground "#c5c8c6") t)))                     ; title
  "For use with the `doom-tomorrow-night' theme.")

(defun b/lights-on ()
  "Enable my favourite light theme."
  (interactive)
  (mapc #'disable-theme custom-enabled-themes)
  (load-theme 'tangomod t)
  (when (featurep 'smart-mode-line)
    (sml/apply-theme 'tangomod))
  (font-lock-remove-keywords
   'org-mode b/org-mode-font-lock-keywords)
  (when (featurep 'erc-hl-nicks)
    (erc-hl-nicks-reset-face-table))
  (when (featurep 'exwm-systemtray)
    (exwm-systemtray--refresh)))

(defun b/lights-off ()
  "Go dark."
  (interactive)
  (mapc #'disable-theme custom-enabled-themes)
  (load-theme 'doom-one t)
  (when (featurep 'smart-mode-line)
    (sml/apply-theme 'automatic))
  (font-lock-add-keywords
   'org-mode b/org-mode-font-lock-keywords t)
  (when (featurep 'erc-hl-nicks)
    (erc-hl-nicks-reset-face-table))
  (when (featurep 'exwm-systemtray)
    (exwm-systemtray--refresh)))

(bind-keys
 ("C-c t d" . b/lights-off)
 ("C-c t l" . b/lights-on))


;;; Emacs enhancements & auxiliary packages

(use-package man
  :config (setq Man-width 80))

(use-package which-key
  :defer 0.4
  :config
  (which-key-add-key-based-replacements
    ;; prefixes for global prefixes and minor modes
    "C-c @"   "outline"
    "C-c !"   "flycheck"
    "C-x RET" "coding system"
    "C-x 8"   "unicode"
    "C-x @"   "event modifiers"
    "C-x a"   "abbrev/expand"
    "C-x r"   "rectangle/register/bookmark"
    "C-x t"   "tabs"
    "C-x v"   "version control"
    "C-x X"   "edebug"
    "C-x C-a" "edebug"
    "C-x C-k" "kmacro"
    ;; prefixes for my personal bindings
    "C-c &"   "yasnippet"
    "C-c a"   "applications"
    "C-c a e" "erc"
    "C-c a o" "org"
    "C-c a s" "shells"
    "C-c b"   "buffers"
    "C-c c"   "compile-and-comments"
    "C-c e"   "eval"
    "C-c f"   "files"
    "C-c F"   "frames"
    "C-c g"   "magit"
    "C-S-h"   "help(ful)"
    "C-c m"   "multiple-cursors"
    "C-c p"   "projectile"
    "C-c p s" "projectile/search"
    "C-c p x" "projectile/execute"
    "C-c p 4" "projectile/other-window"
    "C-c q"   "boxquote"
    "C-c t"   "themes"
    ;; "s-O"     "outline"
    )

  ;; prefixes for major modes
  (which-key-add-major-mode-key-based-replacements 'message-mode
    "C-c f n" "footnote")
  (which-key-add-major-mode-key-based-replacements 'org-mode
    "C-c C-v" "org-babel")

  (which-key-mode)
  :custom
  (which-key-add-column-padding 5)
  (which-key-max-description-length 32))

(use-package crux            ; results in Waiting for git... [2 times]
  :defer 0.4
  :bind (("C-c d"   . crux-duplicate-current-line-or-region)
         ("C-c M-d" . crux-duplicate-and-comment-current-line-or-region)
         ("C-c f C" . crux-copy-file-preserve-attributes)
         ("C-c f D" . crux-delete-file-and-buffer)
         ("C-c f R" . crux-rename-file-and-buffer)
         ("C-c j"   . crux-top-join-line)
         ("C-S-j"   . crux-top-join-line)))

(use-package projectile
  :disabled
  :defer 0.5
  :bind-keymap ("C-c p" . projectile-command-map)
  :config
  (projectile-mode)

  (defun b/projectile-mode-line-fun ()
  "Report project name and type in the modeline."
  (let ((project-name (projectile-project-name))
        (project-type (projectile-project-type)))
    (format "%s%s"
            projectile-mode-line-prefix
            (if project-type
                (format ":%s" project-type)
              ""))))
  (setq projectile-mode-line-function 'b/projectile-mode-line-fun)

  (defun my-projectile-invalidate-cache (&rest _args)
    ;; ignore the args to `magit-checkout'
    (projectile-invalidate-cache nil))

  (eval-after-load 'magit-branch
    '(progn
       (advice-add 'magit-checkout
                   :after #'my-projectile-invalidate-cache)
       (advice-add 'magit-branch-and-checkout
                   :after #'my-projectile-invalidate-cache)))
  :custom
  (projectile-completion-system 'ivy)
  (projectile-mode-line-prefix " proj"))

(use-package helpful
  :defer 0.6
  :bind
  (("C-S-h c" . helpful-command)
   ("C-S-h f" . helpful-callable)        ; helpful-function
   ("C-S-h v" . helpful-variable)
   ("C-S-h k" . helpful-key)
   ("C-S-h p" . helpful-at-point)))

(use-package unkillable-scratch
  :defer 0.6
  :config
  (unkillable-scratch 1)
  :custom
  (unkillable-buffers '("^\\*scratch\\*$" "^\\*Messages\\*$")))

;; ,----
;; | make pretty boxed quotes like this
;; `----
(use-package boxquote
  :defer 0.6
  :bind
  (:prefix-map b/boxquote-prefix-map
               :prefix "C-c q"
               ("b"   . boxquote-buffer)
               ("B"   . boxquote-insert-buffer)
               ("d"   . boxquote-defun)
               ("F"   . boxquote-insert-file)
               ("hf"  . boxquote-describe-function)
               ("hk"  . boxquote-describe-key)
               ("hv"  . boxquote-describe-variable)
               ("hw"  . boxquote-where-is)
               ("k"   . boxquote-kill)
               ("p"   . boxquote-paragraph)
               ("q"   . boxquote-boxquote)
               ("r"   . boxquote-region)
               ("s"   . boxquote-shell-command)
               ("t"   . boxquote-text)
               ("T"   . boxquote-title)
               ("u"   . boxquote-unbox)
               ("U"   . boxquote-unbox-region)
               ("y"   . boxquote-yank)
               ("M-q" . boxquote-fill-paragraph)
               ("M-w" . boxquote-kill-ring-save)))

(use-package orgalist
  ;; breaks auto-fill-mode, showing this error:
  ;; orgalist--boundaries: Lisp nesting exceeds ‘max-lisp-eval-depth’
  :disabled
  :after message
  :hook (message-mode . orgalist-mode))

;; highlight TODOs in buffers
(use-package hl-todo
  :defer 0.5
  :config
  (global-hl-todo-mode))

(use-package multi-term
  :disabled
  :defer 0.6
  :bind (("C-c a s m m" . multi-term)
         ("C-c a s m d" . multi-term-dedicated-toggle)
         ("C-c a s m p" . multi-term-prev)
         ("C-c a s m n" . multi-term-next)
         :map term-mode-map
         ("C-c C-j" . term-char-mode))
  :config
  (setq multi-term-program "screen"
        multi-term-program-switches (concat "-c"
                                            (getenv "XDG_CONFIG_HOME")
                                            "/screen/screenrc")
        ;; TODO: add separate bindings for connecting to existing
        ;; session vs. always creating a new one
        multi-term-dedicated-select-after-open-p t
        multi-term-dedicated-window-height 20
        multi-term-dedicated-max-window-height 30
        term-bind-key-alist
        '(("C-c C-c" . term-interrupt-subjob)
          ("C-c C-e" . term-send-esc)
          ("C-c C-j" . term-line-mode)
          ("C-k" . kill-line)
          ;; ("C-y" . term-paste)
          ("C-y" . term-send-raw)
          ("M-f" . term-send-forward-word)
          ("M-b" . term-send-backward-word)
          ("M-p" . term-send-up)
          ("M-n" . term-send-down)
          ("M-j" . term-send-raw-meta)
          ("M-y" . term-send-raw-meta)
          ("M-/" . term-send-raw-meta)
          ("M-0" . term-send-raw-meta)
          ("M-1" . term-send-raw-meta)
          ("M-2" . term-send-raw-meta)
          ("M-3" . term-send-raw-meta)
          ("M-4" . term-send-raw-meta)
          ("M-5" . term-send-raw-meta)
          ("M-6" . term-send-raw-meta)
          ("M-7" . term-send-raw-meta)
          ("M-8" . term-send-raw-meta)
          ("M-9" . term-send-raw-meta)
          ("<C-backspace>" . term-send-backward-kill-word)
          ("<M-DEL>" . term-send-backward-kill-word)
          ("M-d" . term-send-delete-word)
          ("M-," . term-send-raw)
          ("M-." . comint-dynamic-complete))
        term-unbind-key-alist
        '("C-z" "C-x" "C-c" "C-h"
          ;; "C-y"
          "<ESC>")))

(use-package page-break-lines
  :defer 0.5
  :custom
  (page-break-lines-max-width fill-column)
  :config
  (global-page-break-lines-mode))

(use-package expand-region
  :bind ("C-=" . er/expand-region))

(use-package multiple-cursors
  :bind
  (("C-S-<mouse-1>" . mc/add-cursor-on-click)
   (:prefix-map b/mc-prefix-map
               :prefix "C-c m"
               ("c" . mc/edit-lines)
               ("n" . mc/mark-next-like-this)
               ("p" . mc/mark-previous-like-this)
               ("a" . mc/mark-all-like-this))))

(use-package yasnippet
  :defer 0.6
  :config
  (defconst yas-verbosity-cur yas-verbosity)
  (setq yas-verbosity 2)
  (add-to-list 'yas-snippet-dirs "~/src/git/guix/etc/snippets" t)
  (yas-reload-all)
  (setq yas-verbosity yas-verbosity-cur)

  (defun b/yas--maybe-expand-key-filter (cmd)
    (when (and (yas--maybe-expand-key-filter cmd)
               (not (bound-and-true-p git-commit-mode)))
      cmd))
  (defconst b/yas-maybe-expand
    '(menu-item "" yas-expand :filter b/yas--maybe-expand-key-filter))
  (define-key yas-minor-mode-map
    (kbd "SPC") b/yas-maybe-expand)

  (yas-global-mode))

(use-package debbugs
  :bind
  (("C-c D d" . debbugs-gnu)
   ("C-c D b" . debbugs-gnu-bugs)
   ("C-c D e" .
    (lambda ()
      (interactive)                     ; bug-gnu-emacs
      (setq debbugs-gnu-current-suppress t)
      (debbugs-gnu debbugs-gnu-default-severities '("emacs"))))
   ("C-c D g" .                         ; bug-gnuzilla
    (lambda ()
      (interactive)
      (setq debbugs-gnu-current-suppress t)
      (debbugs-gnu debbugs-gnu-default-severities '("gnuzilla"))))
   ("C-c D G b" .                       ; bug-guix
    (lambda ()
      (interactive)
      (setq debbugs-gnu-current-suppress t)
      (debbugs-gnu debbugs-gnu-default-severities '("guix"))))
   ("C-c D G p" .                       ; guix-patches
    (lambda ()
      (interactive)
      (setq debbugs-gnu-current-suppress t)
      (debbugs-gnu debbugs-gnu-default-severities '("guix-patches"))))))

(use-package org-ref
  :init
  (b/setq-every '("~/usr/org/references.bib")
    reftex-default-bibliography
    org-ref-default-bibliography)
  (setq
   org-ref-bibliography-notes "~/usr/org/notes.org"
   org-ref-pdf-directory "~/usr/org/bibtex-pdfs/"))

;; (use-package fill-column-indicator)

(use-package window
  :bind
  (("C-c w s l" . (lambda ()
                    (interactive)
                    (split-window-right)
                    (other-window 1)))
   ("C-c w s j" . (lambda ()
                    (interactive)
                    (split-window-below)
                    (other-window 1)))
   ("C-c w q"   . quit-window))
  :custom
  (split-width-threshold 150))

(use-package windmove
  :defer 0.6
  :bind
  (("C-c w h" . windmove-left)
   ("C-c w j" . windmove-down)
   ("C-c w k" . windmove-up)
   ("C-c w l" . windmove-right)
   ("C-c w H" . windmove-swap-states-left)
   ("C-c w J" . windmove-swap-states-down)
   ("C-c w K" . windmove-swap-states-up)
   ("C-c w L" . windmove-swap-states-right)))

(use-package pass
  :commands pass
  :bind ("C-c a p" . pass)
  :hook (pass-mode . View-exit))

(use-package pdf-tools
  :defer 0.5
  :bind (:map pdf-view-mode-map
              ("<C-XF86Back>"    . pdf-history-backward)
              ("<mouse-8>"       . pdf-history-backward)
              ("<drag-mouse-8>"  . pdf-history-backward)
              ("<C-XF86Forward>" . pdf-history-forward)
              ("<mouse-9>"       . pdf-history-forward)
              ("<drag-mouse-9>"  . pdf-history-forward)
              ("M-RET"           . image-previous-line)
              ("C-s"             . isearch-forward)
              ("s s"             . isearch-forward))
  :config (pdf-tools-install nil t)
  :custom (pdf-view-resize-factor 1.05))

(use-package org-pdftools
  :disabled
  :straight (:host github :repo "fuxialexander/org-pdftools")
  :demand
  :after org
  :config
  (with-eval-after-load 'org
    (require 'org-pdftools)))

(use-package biblio)

(use-package reftex
  :hook (latex-mode . reftex-mode))

(use-package reftex-cite
  :after reftex
  :disabled                             ; enable to disable
                                        ; reftex-cite's default choice
                                        ; of previous word
  :config
  (defun reftex-get-bibkey-default ()
    "If the cursor is in a citation macro, return the word before the macro."
    (let* ((macro (reftex-what-macro 1)))
      (save-excursion
        (when (and macro (string-match "cite" (car macro)))
          (goto-char (cdr macro)))
        (reftex-this-word)))))

(use-package minions
  :demand
  :config (minions-mode))

(use-package dmenu
  :custom
  (dmenu-prompt-string "run: ")
  (dmenu-save-file (b/var "dmenu-items")))

(use-package eosd
  ;; TODO: fix build by properly building the eosd-pixbuf.c module
  ;; e.g. see https://github.com/raxod502/straight.el/issues/386
  :disabled
  :straight (:host github :repo "clarete/eosd")
  :demand
  :after exwm
  :config
  (eosd-start))

(use-package eww
  :bind ("C-c a e w" . eww)
  :custom
  (eww-download-directory (file-name-as-directory
                           (getenv "XDG_DOWNLOAD_DIR"))))


;;; Post initialization

)
(message "Loading %s...done (%.3fs)" user-init-file
         (float-time (time-subtract (current-time)
                                    b/before-user-init-time)))

;;; init.el ends here
